from athenaSQL.column import WindowColumn

from athenaSQL.functions.functions import _create_window_function

_window_functions = {
    'cume_dist': 'Returns the cumulative distribution of a value in a group of values',
    'dense_rank': 'Returns the rank of a value in a group of values',
    'percent_rank': 'Returns the percentage ranking of a value in group of values',
    'rank': 'Returns the rank of a value in a group of values',
    'row_number': 'Returns a unique sequential number for each row'
}

for _name, _doc in _window_functions.items():
    globals()[_name] = _create_window_function(_name, _doc)

# ntile(n)


def ntile(n):
    """
    Divides the rows for each window partition into n buckets ranging 
    from 1 to at most n.
    """

    if not isinstance(n, int):
        raise TypeError('`{type(n).__name__}` is not of a type `int`')

    wCol = WindowColumn('ntile', n)
    return wCol

# first_value(x)


def first_value(x):
    """
    """
    wCol = WindowColumn('first_value', x)
    return wCol

# last_value(x)


def last_value(x):
    """
    """
    wCol = WindowColumn('last_value', x)
    return x

#nth_value(x, offset)


def nth_value(x, offset):
    """
    """
    wCol = WindowColumn('nth_value', x, offset)
    return wCol

# lead(x[, offset[, default_value]])


def lead(x, offset=None, default_value=None):
    """
    """
    if default_value and not offset:
        raise ValueError('`default_value` is provided without `offset`')

    optional_args = [arg for arg in [offset, default_value] if arg]

    wCol = WindowColumn('lead', x, *optional_args)

    return wCol

# lag(x[, offset[, default_value]])


def lag(x, offset, default_value):
    """
    """
    if default_value and not offset:
        raise ValueError('`default_value` is provided without `offset`')

    optional_args = [arg for arg in [offset, default_value] if arg]

    wCol = WindowColumn('lag', x, *optional_args)

    return wCol


# Ensure that Sphinx finds the dynamically created functions
__all__ = [
    *_window_functions.keys(),
    'ntile',
    'first_value',
    'last_value',
    'nth_value',
    'lead',
    'lag'
]




from typing import Union

class ColumnType:
    """
    column type class, used to identify column types 
    """
    def __init__(self, col_type: str):
        self.col_type: str = col_type

        self._syntax: str = col_type 

    def __str__(self) -> str:
        return self._syntax

class UnaryColumnType(ColumnType):
    """
    column type that takes single operand like varchar(10).
    instantiated with column type as input.
    instance is callable with operand as input.
    """
    def __init__(self, col_type: str):
        super().__init__(col_type)

    def __call__(self, operand: Union[str, int]):
        self._syntax: str = f'{self.col_type}({operand})'
        return self

class BinaryColumnType(ColumnType):
    """
    column type that takes two operands like decimal(10, 25)
    instantiated with column type as input
    instance is ballable with operands as inputs 
    """
    def __init__(self, col_type: str):
        super().__init__(col_type)

    def __call__(self, operand1: int, operand2: int):
        self._syntax: str = f'{self.col_type}({operand1}, {operand2})'
        return self

class _COLUMNTYPE:
    boolean: ColumnType = ColumnType('boolean')
    tinyint: ColumnType = ColumnType('tinyint')
    smallint: ColumnType = ColumnType('smallint')
    int: ColumnType = ColumnType('int')
    bigint: ColumnType = ColumnType('bigint')
    double: ColumnType = ColumnType('double')
    float: ColumnType = ColumnType('float')
    string: ColumnType = ColumnType('string')
    binary: ColumnType = ColumnType('binary')
    date: ColumnType = ColumnType('date')
    timestamp: ColumnType = ColumnType('timestamp')
    char: UnaryColumnType = UnaryColumnType('char')
    varchar: UnaryColumnType = UnaryColumnType('varchar')
    decimal: BinaryColumnType = BinaryColumnType('decimal')
    #TODO array, map, struct

COLUMNTYPE = _COLUMNTYPE()



import copy
from jinja2 import Template

from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_uncalled_preceding_method, \
    _check_and_extract_list_or_valid_typed_arguments

from athenaSQL.orders import ASC, DESC
from athenaSQL.column import Column, ConditionalColumn, AliasColumn, \
    AggregateColumn

select_query_template = """
{% if cte_tables %}
WITH{{' '}}
{%- for name, query in cte_tables.items() -%}
{{ name }} AS (
    {{ query._to_sql()|indent(4) }}
){%- if not loop.last -%},{{' '}}{%- endif -%}
{%- endfor -%}
{% endif %}
SELECT
    {%- if columns -%}
    {%- for col in columns %}
    {{col}}
    {%- if not loop.last -%}
    ,
    {%- endif -%}
    {%- endfor -%}
    {% else %}
    *
    {%- endif %}
FROM {% if database -%}"{{database}}".{%- endif %}"{{table}}"
{% if filters -%}
WHERE
    {% for filter in filters -%}
    {{filter}}{{" "}}
    {%- if not loop.last -%}
    AND
    {% endif -%}
    {% endfor -%}
{% endif %}
{% if grouping_cols -%}
GROUP BY{{" "}}
    {%- for col in grouping_cols -%}
    {{col}}
    {%- if not loop.last -%}
    {{", "}}
    {%- endif -%}
    {%- endfor -%}
{% endif %}
{% if group_filters -%}
HAVING
    {% for filter in group_filters -%}
    {{filter}}{{" "}}
    {%- if not loop.last -%}
    AND
    {% endif -%}
    {% endfor -%}
{% endif %}
{% if ordering_cols -%}
ORDER BY{{" "}}
    {%- for col in ordering_cols -%}
    {{col}}
    {%- if not loop.last -%}
    {{", "}}
    {%- endif -%}
    {%- endfor -%}
{% endif %}
{% if limit -%}
LIMIT {{limit}}
{% endif %}
"""


class SelectQuery(QueryABC):
    def __init__(self, table, database=None):
        self._table = table
        self._database = database

        QueryABC.__init__(self, database, table)

        self._cte_tables = {}
        self._columns = []
        self._filters = []
        self._grouping_cols = []
        self._group_filters = []
        self._limit = ''
        self._ordering_cols = []

    def _to_sql(self):
        """
        """

        # exit if select is grouped and a non grouping column is not aggregated
        # select is grouped
        if self._grouping_cols:
            _unagg_cols = [col for col in self._columns
                           # stringify all because built in __eq__ has ben replaced
                           if str(col) not in list(map(str, self._grouping_cols))
                           and type(col) in (Column, str)]

            if _unagg_cols:
                raise SyntaxError(f"'{_unagg_cols[0]}' must be an aggregated "
                                  "expression or inserted in 'groupBy' method")

        return Template(select_query_template).render(
            table=self._table,
            database=self._database,
            cte_tables=self._cte_tables,
            columns=self._columns,
            filters=self._filters,
            grouping_cols=self._grouping_cols,
            group_filters=self._group_filters,
            ordering_cols=self._ordering_cols,
            limit=self._limit
        ).strip()

    def select(self, *cols):
        """
        specify columns to select 

        Parameters
        ----------
        cols: list or str
        """
        # check if select() method has already been called in an instance
        _exit_on_override(self._columns, 'selects', 'select(cols)')

        # check if argument is either a list, str or Column and return list
        # of args else raises method argument error
        arguments = _check_and_extract_list_or_valid_typed_arguments(cols,
                                                                     'select')

        columns = list(map(lambda arg: arg
                           if isinstance(arg, (Column, AliasColumn))
                           else Column(arg),
                           arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._columns = columns

        return clone_obj

    def filter(self, condition):
        """
        Add fiter condition to where clause

        Parameters
        ----------
        condition: ConditionalColumn
            a sql where caluse ConditionalColumn
        """

        if not isinstance(condition, ConditionalColumn):
            raise TypeError('.filter() can only accept a ConditionalColumn')

        clone_obj = copy.deepcopy(self)
        # update the instances filters
        clone_obj._filters.append(condition)

        return clone_obj

    def groupBy(self, *cols):
        """
        """
        # check if groupBy() method has already been called in an instance
        _exit_on_override(self._grouping_cols,
                          'grouping columns',
                          'groupBy(cols)')

        # check if argument is either a list or string and return list of args
        # else raises method argument error
        # TODO type should be of Column not instance of Column
        arguments = _check_and_extract_list_or_valid_typed_arguments(cols,
                                                                     'groupBy')

        grouping_cols = list(map(lambda arg: arg
                                 if isinstance(arg, (Column, AliasColumn))
                                 else Column(arg),
                                 arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._grouping_cols = grouping_cols

        # if select columns are not specified, add grouping cols
        if len(clone_obj._columns) == 0:
            clone_obj._columns = grouping_cols

        return clone_obj

    def filterGroup(self, condition):
        """
        """

        # exit if groupBy hasn't already been called
        _exit_on_uncalled_preceding_method(self._grouping_cols,
                                           'filterGroup',
                                           'groupBy')

        if not isinstance(condition, ConditionalColumn):
            raise TypeError(
                '.filterGroup() can only accept a ConditionalColumn')

        clone_obj = copy.deepcopy(self)
        # update group filters
        clone_obj._group_filters.append(condition)

        return clone_obj

    def agg(self, *aggs):
        """
        Accept one or more AggregateColumn and either replace or add to
        select column
        """

        # check all aggs are types of AggregateColumn
        # else raises method argument error
        arguments = _check_and_extract_list_or_valid_typed_arguments(aggs,
                                                                     'agg', valid_types=(AggregateColumn, AliasColumn))

        # if host of AliasColumn is not AggregateColumn throw exception
        for agg in aggs:
            if (isinstance(agg, AliasColumn) and
                    not isinstance(agg.host, AggregateColumn)):
                raise TypeError('`.agg()` can only accept `AggregateColumn` or'
                                ' it\'s alias.')

        # extract base column name to check cross check with column list
        agg_columns = [col.column for col in arguments]
        # remove column that are aggregated if exist
        filtered_cols = [col for col in self._columns
                         if col.column not in agg_columns]

        clone_obj = copy.deepcopy(self)
        # update columns with agg columns
        clone_obj._columns = filtered_cols + arguments

        return clone_obj

    def limit(self, limit):
        """
        Add limit to the instance query
        A limit lets you specify how many rows to return, it starts from top

        Parameter
        ---------
        limit: int or str 
            integer or numeric string
        """
        # check if limit() method has already been called in an instance
        _exit_on_override(self._limit,
                          'limit',
                          'limit(limit)')

        if not str(limit).isdigit():
            raise TypeError('.limit() can only acccept an integer '
                            'or an integer numeric string')

        clone_obj = copy.deepcopy(self)
        clone_obj._limit = str(limit)

        return clone_obj

    def orderBy(self, *cols):
        """
        Add columns to order the result by

        Parameters
        ----------
        cols: str or list
        """
        # check if orderBy() method has already been called in an instance
        _exit_on_override(self._ordering_cols,
                          'ordering columns',
                          'orderBy(cols)')

        # check if argument is either a list, ASC or DESC and return list of args
        # else raises method argument error
        arguments = _check_and_extract_list_or_valid_typed_arguments(cols,
                                                                     'orderBy', valid_types=(str, ASC, DESC))

        # should also be done in Column types
        ordering_cols = list(map(lambda arg: arg
                                 if isinstance(arg, (ASC, DESC))
                                 else Column(arg),
                             arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._ordering_cols = ordering_cols

        return clone_obj



from typing import Union, Optional

from athenaSQL.column import Column 


def stringify(colOrPrimitive: Union[Column, int, str, float, bool],
              singleQuote: Optional[bool] = True):
    """Return a string representation of `colOrPrimitive`. If `colOrPrimitive`
    is a string it will be enclosed in a pair of single quotes. 

    Args:
        colOrPrimitive (Union[Column, int, str, float, bool]): a column or 
        primitive to stringify.
        singleQuote (Optional[bool]): whether to use single or double quotes 
        for string values. Default is True.

    Raises:
        TypeError: raised if `colOrPrimitive` is not of supported types. 

    Returns:
        str: return string representation of `colOrPrimitive`.
    """
    if not isinstance(colOrPrimitive, (Column, int, str, float, bool)):
        raise TypeError(f'{type(colOrPrimitive).__name__} is not a type of '
                        '(Column, int, str, float, bool)')

    return (f'\'{colOrPrimitive}\'' 
            if isinstance(colOrPrimitive, str) 
            else str(colOrPrimitive))


from .select import SelectQuery
from .insert import InsertQuery
from .create import CreateQuery
from .window import WindowQuery, Window 
from .create_as import CreateAsQuery
from .cte import CTEQuery, withTable
from .unload import UnloadQuery, unload



from enum import Enum

class DataType(Enum):
    """Supported data types enum class.
    """
    BOOLEAN = 'boolean'
    TINYINT = 'tinyint'
    SMALLINT = 'smallint'
    INT = 'int'
    INTEGER = 'integer'
    BIGINT = 'bigint'
    DOUBLE = 'double'
    FLOAT = 'float'
    CHAR = 'char'
    VARCHAR = 'varchar'
    STRING = 'string'
    BINARY = 'binary'
    DATE = 'date'
    TIMESTAMP = 'timestamp'
    DECIMAL = 'decimal'


from athenaSQL.athena_table import AthenaTable

class Athena:
    """
    An abstraction layer for athena database & table.
    It has a static method that accepts a query string to run agains the whole
    database.

    Parameters
    ----------
    database: str 
        database name to select desired table from 

    Example
    -------
    >>> from athenaSQL.athena import Athena
    >>> cpe_table = Athena('adhouse').table('cpe') # abstraction table class for cpe
    >>> cpe_df = Athena.sql('SELECT * FROM adhouse.cpe') # cpe data in pandas df
    """
    def __init__(self, database: str):
        self.database: str = database

    def table(self, table: str):
        """
        """
        return AthenaTable(database=self.database, table=table)



import copy
from jinja2 import Template

from athenaSQL.queries.select import SelectQuery
from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_partial_query

unload_query_template = """
{% macro render_type(value) -%}
{%- if value is string -%}
'{{ value }}'
{%- elif value is boolean -%}
{{ value|lower }}
{%- else -%}
{{ value }}
{%- endif -%}
{%- endmacro -%}

UNLOAD (
    {{ select_query._to_sql()|indent(4) }}
)
TO '{{ location }}'
{% if tbl_properties -%}
WITH (
    {% for key, prop in tbl_properties.items() -%}
    {%- if key in ['partitioned_by', 'bucketed_by', 'partitioning'] -%}
    {{ key }}=ARRAY[
        {%- for col in prop -%}
        '{{ col }}'
        {%- if not loop.last -%},{{' '}}{%- endif -%}
        {%- endfor -%}
    ]
    {%- else -%}
    {{ key }}={{ render_type(prop) }}
    {%- endif %}
    {%- if not loop.last -%},{%- endif %}
    {% endfor -%}
)
{%- endif -%}
"""


class UnloadQuery(QueryABC):
    """
    """

    def __init__(self):
        QueryABC.__init__(self, None, None)

        self._select_query = None
        self._location = None
        self._tbl_properties = {}

    def _to_sql(self):

        # check if query is compelete before generating sql
        req_att = {'unload': self._select_query,
                   'location': self._location}
        for method, att in req_att.items():
            _exit_on_partial_query(att, method, 'UnloadQuery')

        return Template(unload_query_template).render(
            select_query=self._select_query,
            location=self._location,
            tbl_properties=self._tbl_properties
        ).strip()

    def unload(self, select_query):
        """
        """
        _exit_on_override(self._select_query, 'unload', 'unload(SelectQuery)')

        if not isinstance(select_query, SelectQuery):
            raise TypeError(f'{type(select_query).__name__} is not of a type '
                            f'SelectQuery')

        clone_obj = copy.deepcopy(self)
        clone_obj._select_query = select_query

        return clone_obj

    def location(self, location):
        """
        """
        _exit_on_override(self._location, 'location', 'location(loc)')

        # TODO validate s3 location
        clone_obj = copy.deepcopy(self)
        clone_obj._location = location

        return clone_obj

    def withTblProps(self, props):
        """
        """
        _exit_on_override(self._tbl_properties,
                          'table properties',
                          'withTblProp(props)')

        for k in ['partitioned_by', 'bucketed_by', 'partitioning']:
            if k in props.keys() and not isinstance(props[k], dict):
                TypeError(f'Unsupported type `{type(props[k]).__name__}` for '
                          f'`{k}`, use `dict` type instead.')

        clone_obj = copy.deepcopy(self)
        clone_obj._tbl_properties = props

        return clone_obj


def unload(select_query):
    """
    """
    return UnloadQuery().unload(select_query)



import copy
from jinja2 import Template

from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_partial_query, \
    _check_and_extract_list_or_valid_typed_arguments

from athenaSQL.column import Column
from athenaSQL.queries.select import SelectQuery

insert_query_template = """
INSERT INTO "{{database}}"."{{table}}"
{%- if columns -%}
(
        {%- for col in columns %}
        {{col}}
        {%- if not loop.last -%}
        ,
        {%- endif -%}
        {%- endfor %}
    )
{%- endif %}
{{select_query}}
"""


class InsertQuery(QueryABC):
    def __init__(self, database, table):
        """
        """
        self._database = database
        self._table = table

        QueryABC.__init__(self, database, table)

        self._select_query = None
        self._column_order = []

    def _to_sql(self):
        """
        """

        # required attributes
        req_att = {'insert': self._select_query}

        # check if query constraction is complete
        for method, att in req_att.items():
            _exit_on_partial_query(att, method, 'InsertQuery')

        return Template(insert_query_template).render(
            database=self._database,
            table=self._table,
            columns=self._column_order,
            select_query=self._select_query._to_sql()
        ).strip()

    def insert(self, select_query):
        """
        """
        # check if method is already been called in an instance
        _exit_on_override(self._select_query, 'insert', 'insert(SelectQuery)')

        #
        arguments = _check_and_extract_list_or_valid_typed_arguments(
            select_query,
            'insert',
            valid_types=SelectQuery)

        clone_obj = copy.deepcopy(self)
        clone_obj._select_query = arguments

        return clone_obj

    def column_order(self, *cols):
        """
        """

        #
        _exit_on_override(self._column_order, 'column_order',
                          'column_order(cols)')

        #
        arguments = _check_and_extract_list_or_valid_typed_arguments(
            cols,
            'column_order',
            valid_types=(str, Column)
        )

        # convert string to column type
        columns = list(map(lambda arg: Column(str(arg)), arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._column_order = columns

        return clone_obj



from athenaSQL.column import Column

from athenaSQL.functions.functions import _create_unary_function, \
    _create_binary_function

from athenaSQL.queries.query_abc import _check_and_extract_list_or_valid_typed_arguments

_unary_functions = {
    'chr': 'Returns the Unicode code point n as a single character string.',
    'codepoint': 'Returns the Unicode code point of the only character of string.',
    'length': 'Returns the length of string in characters.',
    'upper': 'Converts a string expression to upper case.',
    'lower': 'Converts a string expression to upper case.',
    'ltrim': 'Removes leading whitespace from string.',
    'reverse': 'Returns string with the characters in reverse order.',
    'rtrim': 'Removes trailing whitespace from string.',
    'trim': 'Removes leading and trailing whitespace from string.',
    'word_stem': 'Returns the stem of word in the English language.'
}

_binary_functions = {
    'hamming_distance': 'Returns the Hamming distance of string1 and string2, '
                        'i.e. the number of positions at which the corresponding '
                        'characters are different. Note that the two strings '
                        'must have the same length.',
    'levenshtein_distance': 'Returns the Levenshtein edit distance of string1 '
                            'and string2, i.e. the minimum number of single-character '
                            'edits (insertions, deletions or substitutions) needed '
                            'to change string1 into string2.'
}


for _name, _doc in _unary_functions.items():
    globals()[_name] = _create_unary_function(_name, _doc)

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc)


def coalesce(*cols):
    """
    """

    arguments = _check_and_extract_list_or_valid_typed_arguments(cols,
                                                                 'coalesce', valid_types=(str, Column))
    
    column = Column('_')
    column._sql_clause = f'COALESCE(' 
    for index, arg in enumerate(arguments, start=1):
        column._sql_clause += (f'\'{arg}\''
                               if isinstance(arg, str)
                               else f'{arg}')

        if index != len(arguments):
            column._sql_clause += ', '

    column._sql_clause += ')'

    return column
    
def concat(*cols):
    """
    """

    arguments = _check_and_extract_list_or_valid_typed_arguments(cols,
                                                                 'concat', valid_types=(str, Column))

    column = Column('_')
    column._sql_clause = f'CONCAT('
    for index, arg in enumerate(arguments, start=1):
        column._sql_clause += (f'\'{arg}\''
                               if isinstance(arg, str)
                               else f'{arg}')

        if index != len(arguments):
            column._sql_clause += ', '

    column._sql_clause += ')'

    return column


def replace(col, search, replace=None):
    """Replaces all instances of search with replace in col.
    """

    if not isinstance(col, (Column, str)):
        raise TypeError('col argument in replace should be a type of '
                        '[Column|str]')

    if isinstance(col, str):
        col = Column(col)

    # TODO type check search and replace

    col._sql_clause = f"REPLACE({col}, '{search}'"
    col._sql_clause += f", '{replace}')" if replace!=None else ")"

    return col


def substring(col, start, length=None):
    """
    Returns a substring from string of length length from the starting 
    position start.Positions start with 1. A negative starting position 
    is interpreted as being relative to the end of the string.
    """

    if not isinstance(col, (Column, str)):
        raise TypeError('col argument in replace should be a type of '
                        '[Column|str]')

    if isinstance(col, str):
        col = Column(col)

    # TODO type check 'start' and 'length'

    col._sql_clause = f"SUBSTRING({col}, {start}"
    col._sql_clause += f", {length})" if length else ")"

    return col


# Ensure that Sphinx finds the dynamically created functions
__all__ = [
    *_unary_functions.keys(),
    *_binary_functions.keys(),
    'coalesce',
    'concat',
    'replace',
    'substring'
]



from athenaSQL.column import Column, NewColumn, \
    ConditionalColumn, CaseColumn, WindowColumn

from athenaSQL.utils import stringify

def ifTrue(condition, true_value, false_value=None):

    if not isinstance(condition, ConditionalColumn):
        raise TypeError(f'{condition} argument is not type of a `Column`')
    
    _false_value = isinstance(false_value, Column) or false_value != None

    condition._sql_clause = (f'IF({condition}, {true_value}'
                             f', {stringify(false_value)})' if _false_value != None else ')')

    return condition

def tryOrNull(expression):
    """Evaluate an expression and handle certain types of errors by returning NULL.
    The sql function identifier is `TRY(expression)`.

    Args:
        expression Column: a sql expression

    Raises:
        TypeError: If expression is not type of Column will raise TypeError

    Returns:
        Column: expression wrapped with `TRY` function
    """
    if not isinstance(expression, Column):
        raise TypeError(f'{expression} argument is not type of a `Column`')
    
    expression._sql_clause = (f'TRY({expression})')
    
    return expression


# Ensure that Sphinx finds the dynamically created functions
__all__ = [
    'ifTrue',
    'tryOrNull'
]


import re
import copy
from typing import Union, Optional
from athenaSQL.column_type import ColumnType
from athenaSQL.data_type import DataType
from athenaSQL.operator_mixin import ComparisonMixin, \
    ArithmeticMixin, \
    LogicalMixin

def validate_column_name(column: str) -> str:
    """
    https://docs.aws.amazon.com/athena/latest/ug/tables-databases-columns-names.html
    """

    # column contains special characters other than underscore
    if not re.match(r'^[A-Za-z0-9_*]+$', column):
        raise ValueError(
            f'{column} is not a valid column name. '
            'Only alphanumeric & underscore (_) are supported.')

    # convert column name to lower case. All column in Athena are lowercase
    column: str = column.lower()

    return column


def areinstances(values: any, _type: type) -> bool:
    if not isinstance(values, (list, tuple)):
        values: list[any] = [values]
    return all(map(lambda v: isinstance(v, _type), values))

class Column(ComparisonMixin, ArithmeticMixin, LogicalMixin):
    """
    Athena column abstraction class.

    Parameters
    ----------
    col: str
        desired column name
    """

    def __init__(self, col: str):
        self.column: str = validate_column_name(col)

        self._sql_clause: str = col

    def __str__(self) -> str:
        return f'{self._sql_clause}'

    def isin(self, *cols):
        """
        """
        if len(cols) == 1 and isinstance(cols[0], (list, set)):
            cols = cols[0]

        self._sql_clause = f'{self._sql_clause} IN ('
        for col in cols:
            if isinstance(col, str):
                self._sql_clause = f"{self._sql_clause}'{col}', "
                continue

            self._sql_clause = f'{self._sql_clause}{col}, '

        # [:-2] remove the last ,
        self._sql_clause = f'{self._sql_clause[:-2]})'

        return ConditionalColumn(self)

    def isNull(self):
        """
        """
        self._sql_clause = f'{self._sql_clause} IS NULL'
        return ConditionalColumn(self)

    def isNotNull(self):
        """
        """
        self._sql_clause = f'{self._sql_clause} IS NOT NULL'
        return ConditionalColumn(self)

    def alias(self, alias: str):
        """
        Add alias for a column.

        Example
        -------
        >>> Column('age') + 10 
        age + 10
        >>> (Column('age') + 10).alias('age10')
        age + 10 AS age10
        """
        # cast alias to string
        alias: str = str(alias)

        self._sql_clause += f' AS {alias}'

        return AliasColumn(alias, host=self)

    def cast(self, dataType: DataType, _try: Optional[bool] = False):
        """Cast column to a given data type.

        Args:
            dataType (DataType): An enum of supported data type to cast to.
            _try (Optional[bool]): Opiton to use `TRY_CAST` instead of `CAST`. Defaults to False.
        
        Raises:
            TypeError: raised if `dataType` is not a type of DataType.

        Returns:
            Column: return `self` for further method chaining.
        """

        # lowercase dataType to check if it is supported
        # dataType = dataType.lower()

        # supported athena data types
        # https://docs.aws.amazon.com/athena/latest/ug/data-types.html
        # supported_types = ['boolean', 'tinyint', 'smallint', 'int', 'integer',
                        #    'bigint', 'double', 'float', 'char', 'varchar',
                        #    'string', 'binary', 'date', 'timestamp', 'decimal']

        if not isinstance(dataType, DataType):
            raise TypeError(f'`{type(dataType).__name__}` is not a type of `DataType`')

        if _try:
            self._sql_clause = f'TRY_CAST({self._sql_clause} AS {dataType.value.upper()})'
        else:
            self._sql_clause = f'CAST({self._sql_clause} AS {dataType.value.upper()})'

        clone_obj: Column = copy.deepcopy(self)
        return clone_obj

    def between(self, lowerBound: Union[int, str], upperBound: Union[int, str]):
        """
        """
        # TODO check if params are int or str, import from other modules
        if not (areinstances([lowerBound, upperBound], str) or
                areinstances([lowerBound, upperBound], int)):
            raise TypeError('unsupported mixed types, '
                            'supported types are [str] or [int]')

        self._sql_clause = (f'{self._sql_clause} BETWEEN {stringify(lowerBound)}'
                            f' AND {stringify(upperBound)}')

        return ConditionalColumn(self)

    def when(self, condition: any, value: any):
        """
        A method to give an exception instraction how to use when condition.
        """
        raise ValueError('when() can only be applied on a Column previously '
                         'generated by when() function')

    def otherwise(self, value: any):
        """
        A method to give an exception instraction how to use otherwise()
        condition.
        """
        raise ValueError('otherwise() can only be applied on a Column '
                         'previously generated by when() function')

    def over(self, window):
        """
        """

        if type(self) not in [AggregateColumn, WindowColumn]:
            raise TypeError('`over()` can only be applied to `aggregated` or '
                            '`window` columns.')

        # scope error
        from athenaSQL.queries.window import WindowQuery
        if not isinstance(window, WindowQuery):
            raise ValueError(f'`{type(window).__name__}` is not a type '
                             f'of `WindowColumn`')

        clone_obj: Column = copy.deepcopy(self)
        clone_obj._sql_clause = f'{self._sql_clause} {window._to_sql()}'

        return clone_obj


class NewColumn(Column):
    """
    A column with column name and it's data type.
    used when creating a new table.
    """

    def __init__(self, col_name: str, data_type: ColumnType):
        #TODO change data_type to DataType enum
        super().__init__(col_name)

        if not isinstance(data_type, ColumnType):
            raise TypeError(f'{data_type} is not a type of ColumnType.')

        self.column_type = data_type

        self._sql_clause = f'`{col_name}` {data_type.__str__()}'


class ConditionalColumn(Column):
    """
    """

    def __init__(self, col: Column):
        super().__init__(col.column)

        self._sql_clause = col.__str__()


class FunctionalColumn(Column):
    """
    A function wrapping a column.
    https://trino.io/docs/current/functions.html
    """

    def __init__(self, col: Column):
        super().__init__(col.column)

        self._sql_clause = col.__str__()


class AggregateColumn(FunctionalColumn):
    """https://trino.io/docs/current/functions/aggregate.html
    """

    def __init__(self, col: Column):
        super().__init__(col)


class WindowColumn(Column):
    """https://trino.io/docs/current/functions/window.html
    """

    def __init__(self, func_name: str, *args: list[str, Column]):
        super().__init__('_')
        # take str representation of args and
        # remove leading and trailing parentheses
        self._sql_clause = f'{func_name}({str(list(args))[1:-1]})'


class AliasColumn:
    """
    A column class with alias. It is a return type for Column.alias() function.
    It doesn't support any further chaining of column functionalities.
    """

    def __init__(self, alias: str, *, host):
        # column type of the alias attached to
        # used to identify column type of the host
        self.host = host
        self.column = alias
        self._sql_clause = host.__str__()

    def __str__(self):
        return f'{self._sql_clause}'


class CaseColumn(Column):
    """
    """

    def __init__(self, first_case):
        self.when_cases = [first_case]
        self.else_case = None

        self._sql_clause = self._build_case_sql()

    def _build_case_sql(self):
        """
        """
        case_clause = 'CASE'

        for (condition, value) in self.when_cases:
            case_clause += f'\n\tWHEN {condition} THEN '
            case_clause += (f'\n\t\t\'{value}\'' if isinstance(value, str)
                            else f'\n\t\t{value}')

        if not isinstance(self.else_case, type(None)):
            case_clause += f'\n\tELSE '
            case_clause += (f'\n\t\t\'{self.else_case}\''
                            if isinstance(self.else_case, str)
                            else f'\n\t\t{self.else_case}')

        case_clause += '\nEND'

        return case_clause

    def when(self, condition, value):
        """
        """
        if not isinstance(condition, ConditionalColumn):
            raise TypeError('condition argument in when() should be a '
                            'ConditionalColumn')

        # TODO type checker for value, but value can be wide range of things
        self.when_cases.append((condition, value))
        self._sql_clause = self._build_case_sql()

        clone_obj = copy.deepcopy(self)
        return clone_obj

    def otherwise(self, value):
        """
        """

        # TODO type checker for value
        self.else_case = value

        _column = Column('_')
        _column._sql_clause = self._build_case_sql()

        return _column


def stringify(colOrPrimitive: Union[Column, int, str, float, bool],
              singleQuote: Optional[bool] = True):
    """Return a string representation of `colOrPrimitive`. If `colOrPrimitive`
    is a string it will be enclosed in a pair of single quotes. 

    Args:
        colOrPrimitive (Union[Column, int, str, float, bool]): a column or 
        primitive to stringify.
        singleQuote (Optional[bool]): whether to use single or double quotes 
        for string values. Default is True.

    Raises:
        TypeError: raised if `colOrPrimitive` is not of supported types. 

    Returns:
        str: return string representation of `colOrPrimitive`.
    """
    if not isinstance(colOrPrimitive, (Column, int, str, float, bool)):
        raise TypeError(f'{type(colOrPrimitive).__name__} is not a type of '
                        '(Column, int, str, float, bool)')

    return (f'\'{colOrPrimitive}\'' 
            if isinstance(colOrPrimitive, str) 
            else str(colOrPrimitive))



from athenaSQL.column import ConditionalColumn

from athenaSQL.functions.functions import _create_binary_function


_binary_functions = {
    'regexp_count': 'Returns the number of occurrence of pattern in string.',
    'regexp_extract_all': 'Returns the substring(s) matched by the regular expression pattern in string.',
    'regexp_extract': 'Returns the first substring matched by the regular expression pattern in string.',
    'regexp_position': 'Returns the index of the first occurrence (counting from 1) of pattern in string. Returns -1 if not found.',
    'regexp_replace': 'Removes every instance of the substring matched by the regular expression pattern from string.',
    'regexp_split': 'Splits string using the regular expression pattern and returns an array. Trailing empty strings are preserved.'
}

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc)


_conditioal_binary_functions = {
    'regexp_like': 'Evaluates the regular expression pattern and determines if it is contained within string.'
}


for _name, _doc in _conditioal_binary_functions.items():
    globals()[_name] = _create_binary_function(_name,
                                               _doc,
                                               func_type=ConditionalColumn)

# Ensure that Sphinx finds the dynamically created functions
__all__ = [
    *_binary_functions.keys(),
    *_conditioal_binary_functions.keys()
]


import copy
from jinja2 import Template

from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_partial_query, \
    _check_and_extract_list_or_valid_typed_arguments

from athenaSQL.column import Column, NewColumn

create_query_template = """

{### schema set ###}
{% set set_schema -%}
{%- if columns -%}(
{%- for col in columns %}
    {{ col }}
{%- if not loop.last -%},{%- endif -%}
{%- endfor %}
)
{%- endif -%}
{%- endset %}

{### partition by set ###}
{% set set_partition -%}
{%- if partitions -%}
PARTITIONED BY (
{%- for partition in partitions -%}
{{ partition }}
{%- if not loop.last -%},{%- endif -%}
{%- endfor -%}
)
{%- endif -%}
{%- endset %}

{### clustered by set ###}
{% set set_cluster -%}
{%- if clusters -%}
CLUSTERED BY (
{%- for cluster in clusters -%}
{{ cluster }}
{%- if not loop.last -%},{%- endif -%}
{%- endfor -%}
) INTO {{ buckets }} BUCKETS
{%- endif -%}
{%- endset %}

{### row format set ###}
{% set set_row_format -%}
{%- if row_format -%}
ROW FORMAT SERDE '{{ row_format }}'
{%- endif -%}
{%- endset %}

{### store format set ###}
{% set set_store_format -%}
{%- if store_format -%}
STORED AS {{ store_format }}
{%- elif input_format and output_format -%}
STORED AS INPUTFORMAT
    '{{ input_format }}'
OUTPUTFORMAT
    '{{ output_format }}'
{%- endif -%}
{%- endset %}

{### serde properties set ###}
{% set set_serde_properties -%}
{%- if serde_properties -%}
WITH SERDEPROPERTIES (
    {%- for key, prop in serde_properties.items() %}
    "{{ key }}"="{{ prop }}"
    {%- if not loop.last -%},{%- endif -%}
    {%- endfor %}
)
{%- endif -%}
{%- endset %}

{### location set ###}
{% set set_location -%}
{%- if location -%}
LOCATION '{{ location }}'
{%- endif -%}
{%- endset %}

{### table properties set ###}
{% set set_tbl_properties -%}
{%- if tbl_properties -%}
TBLPROPERTIES (
    {%- for key, prop in tbl_properties.items() %}
    "{{ key }}"="{{ prop }}"
    {%- if not loop.last -%},{%- endif -%}
    {%- endfor %}
)
{%- endif -%}
{%- endset %}

CREATE EXTERNAL TABLE {% if if_not_exists -%}IF NOT EXISTS{%- endif %}
"{{database}}"."{{table}}" {{ set_schema }}
{{ set_partition }}
{{ set_cluster }}
{{ set_row_format }}
{{ set_store_format }}
{{ set_serde_properties }}
{{ set_location }}
{{ set_tbl_properties }}
"""


class CreateQuery(QueryABC):
    def __init__(self, database, table):
        """
        """
        self._database = database
        self._table = table

        QueryABC.__init__(self, database, table)

        self._if_not_exists = True
        self._columns = []
        self._partition_by = []
        self._clustered_by = []
        self._buckets = None
        self._row_format = None
        self._file_format = None
        self._input_format = None
        self._output_format = None
        self._serde_properties = {}
        self._location = None
        self._tbl_properties = {}

    def _to_sql(self):

        # check if query is compelete before generating sql
        req_att = {'columns': self._columns}
        for method, att in req_att.items():
            _exit_on_partial_query(att, method, 'CreateQuery')

        return Template(create_query_template).render(
            database=self._database,
            table=self._table,
            if_not_exists=self._if_not_exists,
            columns=self._columns,
            partitions=self._partition_by,
            clusters=self._clustered_by,
            buckets=self._buckets,
            row_format=self._row_format,
            store_format=self._file_format,
            input_format=self._input_format,
            output_format=self._output_format,
            serde_properties=self._serde_properties,
            location=self._location,
            tbl_properties=self._tbl_properties
        ).strip()

    def ifNotExists(self, opt):
        """
        """

        _exit_on_override(self._if_not_exists,
                          'ifNotExists', 'ifNotExists(bool)')

        if not isinstance(opt, bool):
            raise TypeError(f'`{type(opt).__name__}` is not of a type `bool`')

        clone_obj = copy.deepcopy(self)
        clone_obj._if_not_exists = opt

        return clone_obj

    def columns(self, *new_columns):
        """
        """

        _exit_on_override(self._columns, 'columns', 'columns(new_columns)')

        arguments = _check_and_extract_list_or_valid_typed_arguments(
            new_columns,
            'columns',
            valid_types=(NewColumn)
        )
        clone_obj = copy.deepcopy(self)
        clone_obj._columns = arguments

        return clone_obj

    def partitionBy(self, *cols):
        """
        """
        _exit_on_override(self._partition_by,
                          'partitionBy',
                          'partitionBy(cols)')

        arguments = _check_and_extract_list_or_valid_typed_arguments(
            cols,
            'partitionBy',
            valid_types=(str, Column)
        )

        columns = list(map(lambda arg: Column(str(arg)), arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._partition_by = columns

        return clone_obj

    def clusterBy(self, *cols, buckets=None):
        """
        """
        _exit_on_override(self._clustered_by,
                          'clusterBy',
                          'clusterBy(*cols, buckets)')

        arguments = _check_and_extract_list_or_valid_typed_arguments(
            cols,
            'clusterBy',
            valid_types=(str, Column)
        )

        if not isinstance(buckets, int):
            raise TypeError(f'{buckets} is not a type of an int')

        columns = list(map(lambda arg: Column(str(arg)), arguments))

        clone_obj = copy.deepcopy(self)
        clone_obj._clustered_by = columns
        clone_obj._buckets = buckets

        return clone_obj

    def row_format(self, lib_name):
        """
        https://docs.aws.amazon.com/athena/latest/ug/serde-about.html
        ROW FORMAT DELIMITED - is not supported
        ROW FORMAT SERDE - supported
        """

        _exit_on_override(self._row_format,
                          'row_format',
                          'row_format(lib_name)')

        clone_obj = copy.deepcopy(self)
        clone_obj._row_format = lib_name

        return clone_obj

    def stored_as(self, file_format):
        """
        Specifies the file format for table data. 
        If omitted, TEXTFILE is the default. Options for file_format are:
        SEQUENCEFILE, TEXTFILE, RCFILE, ORC, PARQUET, AVRO, ION
        """
        _exit_on_override((self._file_format or (self._input_format and self._output_format)),
                          'storage format',
                          'stored_as() or stored_as_io()')

        supported_formats = ['SEQUENCEFILE', 'TEXTFILE', 'RCFILE',
                             'ORC', 'PARQUET', 'AVRO', 'ION']
        if file_format.upper() not in supported_formats:
            raise TypeError(f'`{file_format}` is not a supported file format. '
                            f'Supported formats: {supported_formats}')

        clone_obj = copy.deepcopy(self)
        clone_obj._file_format = file_format.upper()

        return clone_obj

    def stored_as_io(self, *, input_format, output_format):
        """
        INPUTFORMAT input_format_classname OUTPUTFORMAT output_format_classname
        """
        _exit_on_override((self._file_format or (self._input_format and self._output_format)),
                          'storage format',
                          'stored_as() or stored_as_io()')

        clone_obj = copy.deepcopy(self)
        clone_obj._input_format = input_format
        clone_obj._output_format = output_format

        return clone_obj

    def serde_properties(self, serdeprops):
        """
        """
        _exit_on_override(self._serde_properties,
                          'serde properties',
                          'serde_properties(serdeprops)')

        clone_obj = copy.deepcopy(self)
        clone_obj._serde_properties = serdeprops

        return clone_obj

    def location(self, loc):
        """
        """
        _exit_on_override(self._location,
                          'location',
                          'location(loc)')

        # TODO check if loc is a valid s3 path

        clone_obj = copy.deepcopy(self)
        clone_obj._location = loc

        return clone_obj

    def tbl_properties(self, props):
        """
        """
        _exit_on_override(self._tbl_properties,
                          'table properties',
                          'tbl_properties(props)')

        clone_obj = copy.deepcopy(self)
        clone_obj._tbl_properties = props

        return clone_obj



from athenaSQL.column import Column

from athenaSQL.functions.functions import _create_unary_function, \
    _create_binary_function

_unary_functions = {
    # Date and Time Functions
    'date': 'This is an alias for CAST(x AS date).',
    'last_day_of_month': 'Returns the last day of the month.',
    'from_iso8601_timestamp': 'Parses the ISO 8601 formatted string into a '
                              'timestamp with time zone.',
    # Date and Time Convenience Extraction Functions
    'day': 'Returns the day of the month from x.',
    'day_of_month': 'This is an alias for day().',
    'day_of_week': 'Returns the ISO day of the week from x. The value ranges '
                   'from 1 (Monday) to 7 (Sunday).',
    'day_of_year': 'Returns the day of the year from x. The value ranges from 1 to 366.',
    'dow': 'This is an alias for day_of_week().',
    'doy': 'This is an alias for day_of_year().',
    'hour': 'Returns the hour of the day from x. The value ranges from 0 to 23.',
    'millisecond': 'Returns the millisecond of the second from x.',
    'minute': 'Returns the minute of the hour from x.',
    'month': 'Returns the month of the year from x.',
    'quarter': 'Returns the quarter of the year from x. The value ranges from '
               '1 to 4.',
    'second': 'Returns the second of the minute from x.',
    'timezone_hour': 'Returns the hour of the time zone offset from timestamp.',
    'timezone_minute': 'Returns the minute of the time zone offset from timestamp.',
    'week': 'Returns the ISO week of the year from x. The value ranges from 1 to 53.',
    'week_of_year': 'This is an alias for week().',
    'year': 'Returns the year from x.',
    'year_of_week': 'Returns the year of the ISO week from x.',
    'yow': 'This is an alias for year_of_week().',
}

_binary_functions = {}

for _name, _doc in _unary_functions.items():
    globals()[_name] = _create_unary_function(_name, _doc)

for _name, _doc in _binary_functions.items():
    globals()[_name] = _create_binary_function(_name, _doc)
    
# supported athena data types
# https://trino.io/docs/current/functions/datetime.html#interval-functions
supported_units = ['millisecond', 'second', 'minute', 'hour', 'day', 
                    'week', 'month', 'quarter', 'year']
    
def date_diff(unit, timestamp1, timestamp2):
    
    # lowercase dataType to check if it is supported
    unit = unit.lower()
    
    if unit not in supported_units:
        raise TypeError(f'unsupported unit {dataType}')
    
    if not isinstance(timestamp1, (Column, str)):
        raise TypeError('timestamp1 argument in date_diff() should be a Column or Str')
    
    if not isinstance(timestamp2, (Column, str)):
        raise TypeError('timestamp2 argument in date_diff() should be a Column or Str')
    
    _col = Column('_')
    _col._sql_clause = f'date_diff(\'{unit}\', {timestamp1}, {timestamp2})'
    
    return _col

def date_trunc(unit, col):
    
    # lowercase dataType to check if it is supported
    unit = unit.lower()
    
    if unit not in supported_units:
        raise TypeError(f'unsupported unit {dataType}')
    
    if not isinstance(col, (Column, str)):
        raise TypeError('col argument in date_trunc() should be a Column or Str')
    
    col._sql_clause = f'date_trunc(\'{unit}\', {col})'
    
    return col

# Ensure that Sphinx finds the dynamically created functions
__all__ = [
    *_unary_functions.keys(),
    *_binary_functions.keys(),
    'date_diff',
    'date_trunc'
]


import copy
from jinja2 import Template

from athenaSQL.queries.query_abc import QueryABC, \
    _exit_on_override, \
    _exit_on_partial_query

from athenaSQL.queries.select import SelectQuery

create_as_query_template = """

{% macro render_type(value) -%}
{%- if value is string -%}
'{{ value }}'
{%- elif value is boolean -%}
{{ value|lower }}
{%- else -%}
{{ value }}
{%- endif -%}
{%- endmacro -%}

CREATE TABLE "{{database}}"."{{table}}"
{% if tbl_properties -%}
WITH (
    {% for key, prop in tbl_properties.items() -%}
    {%- if key in ['partitioned_by', 'bucketed_by', 'partitioning'] -%}
    {{ key }}=ARRAY[
        {%- for col in prop -%}
        '{{ col }}'
        {%- if not loop.last -%},{{' '}}{%- endif -%}
        {%- endfor -%}
    ]
    {%- else -%}
    {{ key }}={{ render_type(prop) }}
    {%- endif %}
    {%- if not loop.last -%},{%- endif %}
    {% endfor -%}
)
{% endif -%}
AS 
{{ select_query._to_sql() }}
{% if not with_data -%}
WITH NO DATA
{%- endif -%}
"""


class CreateAsQuery(QueryABC):
    """
    Creates a new table populated with the results of a SELECT query.
    https://docs.aws.amazon.com/athena/latest/ug/create-table-as.html
    """

    def __init__(self, database, table):
        self._database = database
        self._table = table

        self._tbl_properties = {}
        self._select_query = None
        self._with_data = True

    def _to_sql(self):

        # check if query is compelete before generating sql
        req_att = {'createAs': self._select_query}
        for method, att in req_att.items():
            _exit_on_partial_query(att, method, 'CreateAsQuery')

        return Template(create_as_query_template).render(
            database=self._database,
            table=self._table,
            select_query=self._select_query,
            tbl_properties=self._tbl_properties,
            with_data=self._with_data
        ).strip()

    def withData(self, opt):
        """
        """

        _exit_on_override(self._with_data, 'withData', 'withData(bool)')

        if not isinstance(opt, bool):
            raise TypeError(f'`{type(opt).__name__}` is not of a type `bool`')

        clone_obj = copy.deepcopy(self)
        clone_obj._if_not_exists = opt

        return clone_obj

    def createAs(self, select_query):
        """
        """

        _exit_on_override(self._select_query, 'createAs',
                          'createAs(SelectQuery)')

        if not isinstance(select_query, SelectQuery):
            raise TypeError(f'`{type(select_query).__name__}` is not of a '
                            'type `SelectQuery`')

        clone_obj = copy.deepcopy(self)
        clone_obj._select_query = select_query

        return clone_obj

    def withTblProps(self, props):
        """
        """
        _exit_on_override(self._tbl_properties,
                          'table properties',
                          'withTblProp(props)')

        for k in ['partitioned_by', 'bucketed_by', 'partitioning']:
            if k in props.keys() and not isinstance(props[k], dict):
                TypeError(f'Unsupported type `{type(props[k]).__name__}` for '
                          f'`{k}`, use `dict` type instead.')

        clone_obj = copy.deepcopy(self)
        clone_obj._tbl_properties = props

        return clone_obj
